#include "VCalendar.hpp"

// fBNg
#define DIR_NAME         "/calendar/"
// CfbNXt@C
#define INDEX_FILE_NAME  "index.dat"
// VCalendar ̊gq
#define EXTENSION_NAME   ".vcs"

VCalendar::VCalendar(Void)
{
    return;
}

VCalendar::VCalendar(SFXAnsiStringConstRef string)
{
    Import(string);
    return;
}

VCalendar::VCalendar(SFXAnsiStringConstRef version, SFXDateConstRef startDate,
                     SFXDateConstRef endDate, SFXDateConstRef lastModifiedDate,
                     SFXAnsiStringConstRef summary) : _version(version)
{
    Set(startDate, endDate, lastModifiedDate, summary);
    return;
}

// f[^ݒ
Void VCalendar::Set(SFXDateConstRef startDate, SFXDateConstRef endDate,
    SFXDateConstRef lastModifiedDate, SFXAnsiStringConstRef summary)
{
    _startDate = startDate;
    _endDate = endDate;
    _lastModifiedDate = lastModifiedDate;
    _summary = summary;
    return;
}

// o[Wݒ
Void VCalendar::SetVersion(SFXAnsiStringConstRef version)
{
    _version = version;
    return;
}

SFXDate VCalendar::GetStartDate(Void) const
{
    return _startDate;
}

SFXDate VCalendar::GetEndDate(Void) const
{
    return _endDate;
}

SFXDate VCalendar::GetLastModifiedDate(Void) const
{
    return _lastModifiedDate;
}

SFXAnsiString VCalendar::GetSummary(Void) const
{
    return _summary;
}

// vCalendar`̕荞
Void VCalendar::Import(SFXAnsiStringConstRef string)
{
    SInt32 c1, c2;
    SFXAnsiStringConst format("YYYYMMDD%Thhmmss");
    SFXAnsiString field, value;

    _others.Clear();
    // "BEGIN:VCALENDAR" ݂͑Ȃꍇ́AG[o܂
    c1 = string.FirstIndexOf("BEGIN:VCALENDAR");
    if (c1 < 0) {
        c1 = string.FirstIndexOf("begin:vcalendar");
        if (c1 < 0) {
            return;
        }
    }
    c2 = string.FirstIndexOf('\n', c1);
    while (true) {
        //  ":" 
        c1 = string.FirstIndexOf(':', c2);
        if (c1 < 0) {   // Ȃꍇ
            break;      // ~܂
        }

        // ****:**** ̃R̍擾āǍ̋󔒕̓Jbg
        field = string.Substring(c2 + 1, c1).Trim();

        // ̍sT
        c2 = string.FirstIndexOf('\n', c1);
        if (c2 < 0) {                   // Ȃꍇ
            c2 = string.GetLength();    // ̍ŌƂ
        }

        // ****:**** ̃R̉E擾āǍ̋󔒕̓Jbg
        value = string.Substring(c1 + 1, c2).Trim();

        if (field.Equals("begin", false)) { // 啶𖳎Ĕr
            ; //
        } else if (field.Equals("end", false)){
            if (value.Equals("vcalendar"), false) {
                break; // END:VCALENDAR_ŏI
            } // ȊOȂ疳
        } else if (field.Equals("version", false)) { // o[W
            _version = value;
        } else if (field.Equals("dtstart", false)) { // Jn
            _startDate.Parse(format, value);         // ͂Đݒ
        } else if (field.Equals("dtend", false)) {   // I
            _endDate.Parse(format, value);           // ͂Đݒ
        } else if (field.Equals("last-modified", false)) { // XV
            _lastModifiedDate.Parse(format, value);  // ͂Đݒ
        // Cxgf[^
        } else if (field.Substring(0, 7).Equals("summary", false)) {
            _summary = value;
        } else {    // ̑̃f[^
            _others << field << ":" << value << "\r\n";
        }
    }
    return;
}

// vCalendar`̃f[^oB
Void VCalendar::Export(SFXAnsiStringPtr string) const
{
    SFXAnsiString format("YYYYMMDD%Thhmmss%Z");

    string->Set("BEGIN:VCALENDAR\r\nVERSION:");
    *string += _version;
    *string += "\r\nBEGIN:VEVENT\r\nCLASS:PUBLIC\r\nDTSTART:";
    *string += _startDate.Format(format);
    *string += "\r\nDTEND:";
    *string += _endDate.Format(format);
    *string += "\r\nSUMMARY;ENCODING=QUOTED-PRINTABLE:";
    *string += _summary;
    *string += "\r\nLAST-MODIFIED:";
    *string += _lastModifiedDate.Format(format);
    *string += "\r\n";
    *string += _others;
    *string += "END:VEVENT\r\nEND:VCALENDAR\r\n";
    return;
}

// f[^ӂɎʂ邽߂IDi= t@Cj𔭍sB
// ID̎dl YYYYMMDDHHmmssXXX XXX͂R̗
SFXAnsiString VCalendar::PublishXid(SFXDateConstRef currentDate)
{
    SFXLCGRandom random;
    SFXAnsiString result;

    result = currentDate.Format("YYYYMMDD%Thhmmss");
    result += SFXAnsiString::Format("%03d", random.GetUInt32() % 1000);
    return result;
}

VCalendarFile::VCalendarFile(SFXPathConstRef fileName, SFXAnsiStringConstRef flag)
        : VCalendar(), _fileName(fileName), _flag(flag)
{
    return;
}

VCalendarFile::VCalendarFile(SFXAnsiStringConstRef content,
    SFXPathConstRef fileName, SFXAnsiStringConstRef flag)
        : VCalendar(content), _fileName(fileName), _flag(flag)
{
    return;
}

VCalendarFile::VCalendarFile(SFXAnsiStringConstRef version,
    SFXDateConstRef startDate, SFXDateConstRef endDate,
    SFXAnsiStringConstRef summary) : _flag("-")
{
    SFXDate lastModifiedDate = SFXDate::CurrentDate();
    //t@C last-modified 琶
    _fileName.Set(VCalendar::PublishXid(lastModifiedDate) + EXTENSION_NAME);
    Set(startDate, endDate, lastModifiedDate, summary);
    SetVersion(version);
    return;
}

Void VCalendarFile::SetFileName(SFXPathConstRef fileName)
{
    _fileName = fileName;
    return;
}

Void VCalendarFile::SetFlag(SFXAnsiStringConstRef flag)
{
    _flag = flag;
    return;
}

SFXPath VCalendarFile::GetFileName(Void) const
{
    return _fileName;
}

SFXAnsiString VCalendarFile::GetFlag(Void) const
{
    return _flag;
}

// f[^t@Cɕۑ
SFCError VCalendarFile::SaveToFile(Void) const
{
    SFCError error;
    SFXFile file;
    SFXAnsiStringStreamWriter writer;
    SFXPath fileName(DIR_NAME + _fileName.Get());
    SFXPath temp;
    SFXAnsiString content;

    // ꎞt@C擾
    error = SFXFile::GetTemporaryPath(SFXPath(DIR_NAME), &temp);
    if (error == SFERR_NO_ERROR) {
        // ꎞt@C쐬
        error = file.OpenReadWrite(temp);
        if (error == SFERR_NO_ERROR) {
            // VCalendarNXvCalendar`ɕϊ
            Export(&content);
            // ݃Xg[擾
            error = file.GetStreamWriter(content.GetLength(), &writer);
            if (error == SFERR_NO_ERROR) {
                // t@Cɏ
                writer << content;
                writer.Flush();
            }
            // t@C
            file.Close();
        }
    }
    if (error == SFERR_NO_ERROR) {
        // t@C폜
        SFXFile::Remove(fileName);
        // ꎞt@C{̃t@CɃl[
        error = SFXFile::Rename(temp, fileName);
    }
    return error;
}

// t@Cf[^ǂݏo
SFCError VCalendarFile::LoadFromFile(Void)
{
    SFCError error;
    SFXFile file;
    SFXAnsiStringStreamReader reader;
    UInt32 size;
    SFXAnsiString string;
    SFXPath filename(SFXPath(DIR_NAME + _fileName.Get()));

    // t@C̃TCY擾
    error = SFXFile::GetSize(filename, &size);
    if (error == SFERR_NO_ERROR) {
        // t@CJ
        error = file.OpenReadOnly(filename);
        if (error == SFERR_NO_ERROR) {
            // ǂݍރXg[擾
            error = file.GetStreamReader(size, &reader);
            if (error == SFERR_NO_ERROR) {
                reader.Fetch();
                // t@Cǂݍ
                reader >> string;
            }
            file.Close();
        }
    }
    if (error == SFERR_NO_ERROR) {
        Import(string);         // vCalendar f[^͂Ċeoϐɑ
    }
    return error;
}

VCalCollection::VCalCollection(Void) : _lastSyncTime("00000000T000000Z")
{
    return;
}

VCalCollection::~VCalCollection(Void)
{
    SFXList<VCalendarFilePtr>::Enumerator e = _list.GetFirstEnumerator();
    while(e.HasNext()) {
        VCalendarFilePtr vcal = e.GetNext();
        delete vcal;
    }
    return;
}

Void VCalCollection::SetLastSyncTime(SFXAnsiStringConstRef lastSyncTime)
{
    _lastSyncTime = lastSyncTime;
    return;
}

SFXAnsiString VCalCollection::GetLastSyncTime(Void) const
{
    return _lastSyncTime;
}


// VCalCollection Ƀf[^
Void VCalCollection::Append(VCalendarFilePtr vcalendar)
{
    SFXDateConst date1(vcalendar->GetStartDate());
    Bool isInsert = false;
    SIntN i;

    // VK쐬\tOZbg
    vcalendar->SetFlag("n");

     // f[^}ꏊTB
    // f[^\[gĂKv邽߁B
    for (i = 0; i < _list.GetSize(); ++i) {
        // iԖڂ̃f[^̎擾
        SFXDate date2 = _list.Get(i)->GetStartDate();
        if (date1 <= date2) {   // ̔r
            _list.Insert(i, vcalendar);   // f[^}
            isInsert = true;
            break;
        }
    }

    // }ꏊȂ
    if (!isInsert) {
        _list.InsertLast(vcalendar); // Ōɒǉ
    }

    // f[^t@Cɕۑ
    vcalendar->SaveToFile();
    WriteIndexFile(); // index XV
    return;
}

//f[^̍XVsB
Void VCalCollection::Update(VCalendarFilePtr vcalendar)
{
    _list.Remove(_list.FirstIndexOf(vcalendar));
    Append(vcalendar);
    vcalendar->SetFlag("m");
    return;
}

//f[^ɏtO"d"𗧂Ă
Void VCalCollection::Delete(VCalendarFilePtr vcalendar)
{
    vcalendar->SetFlag("d");
    WriteIndexFile();
    return;
}

//f[^Sɖ
Void VCalCollection::Remove(VCalendarFilePtr vcalendar)
{
    // t@C폜
    SFXFile::Remove(SFXPath(DIR_NAME + vcalendar->GetFileName().Get() + EXTENSION_NAME));
    _list.Remove(_list.FirstIndexOf(vcalendar));
    WriteIndexFile();
    return;
}

// indexԖڂEnumerator擾
SFXList<VCalendarFilePtr>::Enumerator VCalCollection::GetEnumerator(SInt32 index) const
{
    return _list.GetEnumerator(index);
}

VCalendarFilePtr VCalCollection::Get(SInt32 index) const
{
    return _list.Get(index);
}

VCalendarFilePtr VCalCollection::Get(SFXPathConstRef fileName) const
{
    SFXList<VCalendarFilePtr>::Enumerator e = _list.GetFirstEnumerator();
    while(e.HasNext()) {
        VCalendarFilePtr vcal = e.GetNext();
        if (vcal->GetFileName() == fileName) {
            return vcal;
        }
    }
    return null;
}

// yearNAmonthAdayuȍ~v̍ŏ̃f[^indexŎ擾
SInt32 VCalCollection::GetIndex(SInt16 year, SInt16 month, SInt16 day) const
{
    SInt32 i;
    SFXDate date1(year, month, day);
    for (i = 0 ; i < _list.GetSize() ; i++) {
        VCalendarFilePtr vcal = _list.Get(i);
        SFXDate date2 = vcal->GetStartDate();
        date2.SetTime(0, 0, 0);
        if (date1 <= date2) {
            break;
        }
    }
    return i;
}

// yearNAmonthAdaỹf[^̌擾B
// 炩 GetIndex  todayIndex 擾Ă
SInt32 VCalCollection::GetNumberOfVCal(SInt32 todayIndex,
                                    SInt16 year, SInt16 month, SInt16 day) const
{
    SFXDate date1(year, month, day);
    SInt32 size = 0;
    SFXList<VCalendarFilePtr>::Enumerator e = _list.GetEnumerator(todayIndex);
    while(e.HasNext()) {
        VCalendarFilePtr vcal = e.GetNext();
        SFXDate date2 = vcal->GetStartDate();
        date2.SetTime(0, 0, 0);
        if (date1 != date2) {
            break;
        }
        ++size; // tȂ +1 
    }
    return size;
}

// yearNAmonthAdaỹf[^݂Ȃ
Bool VCalCollection::IsDataExist(SInt16 year, SInt16 month, SInt16 day) const
{
    SFXList<VCalendarFilePtr>::Enumerator e =
        _list.GetEnumerator(GetIndex(year, month, day));
    if (e.HasNext()) {
        SFXDate date = e.GetNext()->GetStartDate();
        if (date.GetYear() == year && date.GetMonth() == month
        && date.GetDay() == day) {
            return true;
        }
    }
    return false;
}

// Sf[^t@CǂݍށB
// ܂Aindext@Cǂݍ݁A
// ɓǂݍindext@C͂Aet@C
// VCalCollection Ƀf[^ǂݍށB
SFCError VCalCollection::ReadIndexFile(Void)
{
    SFCError error;
    SFXFile file;
    SFXAnsiStringStreamReader reader;
    UInt32 size;
    SFXAnsiString string;

    //t@C݂Ȃꍇ́A쐬
    bool exists;
    error =SFXFile::Exists( SFXPath(DIR_NAME INDEX_FILE_NAME),&exists);
    if (exists==false)
    {
        //IɃt@C쐬
        file.Create(SFXPath(DIR_NAME INDEX_FILE_NAME),true);
    }

    // t@C̃TCY擾
    error = SFXFile::GetSize(SFXPath(DIR_NAME INDEX_FILE_NAME), &size);
    if (error == SFERR_NO_ERROR) {
        // t@CJ
        error = file.OpenReadOnly(SFXPath(DIR_NAME INDEX_FILE_NAME));
        if (error == SFERR_NO_ERROR) {
            // ǂݍރXg[擾
            error = file.GetStreamReader(size, &reader);
            if (error == SFERR_NO_ERROR) {
                reader.Fetch();
                // t@Cǂݍ
                reader >> string;
            }
            file.Close();
        }
    }
    if (error == SFERR_NO_ERROR) {
        SInt32 c1 = 0;
        SInt32 c2 = 0;
        c2 = string.FirstIndexOf('\n');
        _lastSyncTime = string.Substring(0, c2).Trim();
        ++c2;
        while ((c1 = string.FirstIndexOf(' ', c2)) > -1) {
            SFXPath filename(string.Substring(c2, c1));
            c2 = string.FirstIndexOf("\r\n", c1);
            SFXAnsiString flag(string.Substring(c1 + 1, c2));
            c2 += 2; // \r\n̒

            VCalendarFilePtr vcal = ::new VCalendarFile(filename, flag);
            vcal->LoadFromFile();   // vCalendar f[^t@Cǂݍ
            _list.InsertLast(vcal);
        }
    }
    return error;
}

// t@Cɏ
SFCError VCalCollection::WriteIndexFile(Void)
{
    SFCError error;
    SFXFile file;
    SFXAnsiStringStreamWriter writer;
    SFXPath fileName(DIR_NAME INDEX_FILE_NAME);
    SFXPath temp;
    SFXAnsiString content;

     // ꎞt@C擾
    error = SFXFile::GetTemporaryPath(SFXPath(DIR_NAME), &temp);
    if (error == SFERR_NO_ERROR) {
        // ꎞt@C쐬
        error = file.OpenReadWrite(temp);
        if (error == SFERR_NO_ERROR) {
            // ݃Xg[擾
            error = file.GetStreamWriter(1024, &writer);
            if (error == SFERR_NO_ERROR) {
                // t@Cɏ
                writer << _lastSyncTime << "\r\n";
                writer.Flush();
                SFXList<VCalendarFilePtr>::Enumerator etor
                    = _list.GetFirstEnumerator();
                while(etor.HasNext()) {
                    VCalendarFilePtr vcal = etor.GetNext();
                    writer << vcal->GetFileName().Get() << " "
                        << vcal->GetFlag() << "\r\n";
                    writer.Flush();
                }
            }
            // t@C
            file.Close();
        }
    }
    if (error == SFERR_NO_ERROR) {
        // t@C폜
        SFXFile::Remove(fileName);
        // ꎞt@C{̃t@CɃl[
        error = SFXFile::Rename(temp, fileName);
    }
    return error;
}

SFXAnsiString VCalCollection::MakeListForSync(Void)
{
    SFXAnsiString listStr(_lastSyncTime);
    SFXList<VCalendarFilePtr>::Enumerator e = _list.GetFirstEnumerator();
    while (e.HasNext()) {
        VCalendarFilePtr vcal = e.GetNext();
        if (vcal->GetFlag() != "-") {
            SFXAnsiString content;
            vcal->Export(&content);
            listStr << "####" << vcal->GetFlag() << "#!#!"
                << vcal->GetFileName().Get() << "#!#!" << content << "\r\n";
        }
    }
    return listStr;
}
